/*
 *  configuration.c - functions parsing the configuration file
 *
 *  Copyright (C) 1998 Hugo Haas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <netdb.h>

#include <netinet/ip_icmp.h>
#include "icmp.h"

#include "defines.h"
#include "configuration.h"
#include "filter.h"
#include "netutils.h"

extern struct filter_entry *filter;

/*
 * open_configuration
 *
 * Open the configuration file
 */

FILE *open_configuration(char *name) {
  FILE *configfile;

  configfile = fopen(name, "r");
  if (configfile == NULL) {
    printf("Cannot open configuration file %s.\n", name);
    exit(1);
  }
  return configfile;
}

/*
 * log_protocols
 *
 * Parse the configuration file to find which IP protocols have to be logged
 *
 */

int log_protocols(char *filename) {
  FILE *configfile;
  char line[LINE];
  char *keyword;
  unsigned short result = NONE;

  configfile = open_configuration(filename);
  for(;;) {
    fgets(line, LINE, configfile);
    if (feof(configfile))
      break;
    line[LINE] = '\0';

    keyword = strtok(line, DELIMITERS);
    if (keyword && !strcasecmp(keyword, RUN_KW)) {

      while((keyword = strtok(NULL, DELIMITERS))) {
	if (*keyword == COMMENT_MARK)
	  continue;
	if (!strcasecmp(keyword, ICMP_KW))
	  result |= RUN_ICMP;
	else if (!strcasecmp(keyword, TCP_KW))
	  result |= RUN_TCP;
      }
    }

  }
  fclose(configfile);
  return result;
}

/*
 * get_type
 *
 * Returns a type
 */

long get_port(const char *type, const short protocol) {
  const char *c;
  long port;
 
  /* Number? */
  for (c = type; *c; c++) {
    if (!isdigit(*c)) {
      /* Text */
      switch(protocol) {
      case TCP:
	if (getservbyname(type, "tcp") == NULL)
	  return NOT_DEFINED;
	port = getservbyname(type, "tcp") ->s_port;
	port = ((port << 8) & 0xff00) + ((port >> 8) & 0x00ff);
	return port;
      case ICMP:
	if (!strcasecmp(type,ICMP_ECHOREPLY_KW))
	  return ICMP_ECHOREPLY; 
	else if (!strcasecmp(type, ICMP_DEST_UNREACH_KW))
	  return ICMP_DEST_UNREACH;
	else if (!strcasecmp(type, ICMP_SOURCE_QUENCH_KW))
	  return ICMP_SOURCE_QUENCH;
	else if (!strcasecmp(type, ICMP_REDIRECT_KW))
	  return ICMP_REDIRECT;
	else if (!strcasecmp(type, ICMP_ECHO_KW))
	  return ICMP_ECHO;
	else if (!strcasecmp(type, ICMP_ROUTER_ADVERT_KW))
	  return ICMP_ROUTER_ADVERT;
	else if (!strcasecmp(type, ICMP_ROUTER_SOLICITATION_KW))
	  return ICMP_ROUTER_SOLICITATION;
	else if (!strcasecmp(type, ICMP_TIME_EXCEEDED_KW))
	  return ICMP_TIME_EXCEEDED;
	else if (!strcasecmp(type, ICMP_PARAMETERPROB_KW))
	  return ICMP_PARAMETERPROB;
	else if (!strcasecmp(type, ICMP_TIMESTAMP_KW))
	  return ICMP_TIMESTAMP;
	else if (!strcasecmp(type, ICMP_TIMESTAMPREPLY_KW))
	  return ICMP_TIMESTAMPREPLY;
	else if (!strcasecmp(type, ICMP_INFO_REQUEST_KW))
	  return ICMP_INFO_REQUEST;
	else if (!strcasecmp(type, ICMP_INFO_REPLY_KW))
	  return ICMP_INFO_REPLY;
	else if (!strcasecmp(type, ICMP_ADDRESS_KW))
	  return ICMP_ADDRESS;
	else if (!strcasecmp(type, ICMP_ADDRESSREPLY_KW))
	  return ICMP_ADDRESSREPLY;
	else
	  return NOT_DEFINED;
      default:
	return NOT_DEFINED;
      }
    }
  }

  return atoi(type);
}

/*
 * parse_filter_configuration
 *
 * Parse filter related configuration
 */

void parse_host_options(char *filename) {
  FILE *configfile;
  char *line;
  char *keyword;
  struct filter_entry *new_entry;
  short done_type, done_from;

  char *c;
  struct in_addr inp;

  configfile = open_configuration(filename);
  line = (char *) malloc(LINE);

  for(;;) {

    fgets(line, LINE, configfile);
    if (feof(configfile))
      break;

    keyword = strtok(line, DELIMITERS);
    if (keyword && (!strcasecmp(keyword, LOG_KW) || !strcasecmp(keyword, IGNORE_KW))) {

      new_entry = (struct filter_entry *) malloc(sizeof(struct filter_entry));
      if (new_entry == NULL)
	exit(1);

      /* Log or ignore? */
      if (!strcasecmp(keyword, LOG_KW))
	new_entry->log = TRUE;
      else
	new_entry->log = FALSE;

      /* Protocols */
      keyword = strtok(NULL, DELIMITERS);
      if ((keyword == NULL) || (*keyword == COMMENT_MARK))
	goto cleanup;
      if (!strcasecmp(keyword, ALL_KW))
	new_entry->protocol = ALL;
      else if (!strcasecmp(keyword, ICMP_KW))
	new_entry->protocol = ICMP;
      else if (!strcasecmp(keyword, TCP_KW))
	new_entry->protocol = TCP;

      new_entry->range_min = NOT_DEFINED;
      new_entry->range_max = NOT_DEFINED;
      new_entry->address = 0;
      new_entry->mask = 0;

      done_type = FALSE;
      done_from = FALSE;

      while((keyword = strtok(NULL, DELIMITERS))) {

	if (*keyword == COMMENT_MARK)
	  goto do_register;

	if ((done_type == TRUE) && (done_from == TRUE))
	  goto do_register;

	/* Type description */
	if (!strcasecmp(keyword, TYPE_KW)) {
	  if (done_type == TRUE)
	    continue;
	  keyword = strtok(NULL, DELIMITERS);
          if (*keyword == COMMENT_MARK)
	    goto do_register;
	  /* Range */
	  if ((c = strchr(keyword, RANGE_DM))) {
	    if ((c = keyword)) {
	      new_entry->range_min = NOT_DEFINED;
	      new_entry->range_max = get_port(keyword+1, new_entry->protocol);
	    }
	    else {
	      *c++ = '\0';
	      new_entry->range_min = get_port(keyword, new_entry->protocol);
	      if (*c == '\0')
		new_entry->range_max = NOT_DEFINED;
	      else
		new_entry->range_max = get_port(c, new_entry->protocol);
	      c--;
	      *c = RANGE_DM;
	    }
	  }
	  /* Single number */
	  else {
	    new_entry->range_min = get_port(keyword, new_entry->protocol);
	    new_entry->range_max = new_entry->range_min;
	  }
	  done_type = TRUE;
	}

	/* From description */
	if (!strcasecmp(keyword, FROM_KW)) {
	  if (done_from == TRUE)
	    continue;
	  keyword = strtok(NULL, DELIMITERS);
	  if (*keyword == COMMENT_MARK)
	    goto do_register;
	  /* Address/Mask */
	  if ((c = strchr(keyword, MASK_DM))) {
	    *c++ = '\0';
	    if (!inet_aton(keyword, &inp))
	      goto address_mask_exit;
	    new_entry->address = inp.s_addr;
	    if (c) {
	      if (!inet_aton(c, &inp)) {
		new_entry->address = 0;
		goto address_mask_exit;
	      }
	      new_entry->mask = inp.s_addr;
	    }
	    else
	      new_entry->mask = 0xffffffff;
	  address_mask_exit:
	    *(c-1) = MASK_DM;
	  }
	  else {
	    /* Address */
	    if (isip(keyword)) {
	      if (!inet_aton(keyword, &inp))
		continue;
	      new_entry->address = inp.s_addr;
	      new_entry->mask = 0xffffffff;
	    }
	    else {
	      new_entry->hostname = strdup(keyword);
	      c = new_entry->hostname;
	      while (*c) {
		*c = tolower(*c);
		c++;
	      }
	    }
	  }
	  done_from = TRUE;
	}

      }
    do_register:
      new_entry->next = filter;
      filter = new_entry;
      continue;
    cleanup:
      if (new_entry->hostname != NULL)
	free(new_entry->hostname);
      free(new_entry);
    }

  }

  free(line);
  fclose(configfile);

}

/*
 * $Log: configuration.c,v $
 * Revision 1.4  1998/11/29 21:51:12  hugo
 * Now recognises ICMP message keywords
 *
 * Revision 1.3  1998/11/29 20:48:32  hugo
 * Added configuration parsing for the new configuration file
 *
 * Revision 1.2  1998/10/25 23:19:11  hugo
 * Added parsing for 'from' options
 *
 */
